; We need a serial handler to spend all of it's life reading the serial
; port and filling a queue with the data. This is to overcome serial
; overrun problems when running on slow hardware with fast modems.
;
;   0   long  qend
;   4   long  writeptr
;   8   long  readptr
;   c   long  non-zero if serial error
;  10   char[] start of queue
;
        .text
        .even

        .globl    _qdosjob

_qdosjob:

        move.l  $8(sp),a4               ; struct pointer
        move.l  $4(sp),a5               ; serial input channel
        link    a6,#-64                 ; get some buffer space
loop:
        move.l  a5,a0
        moveq   #3,d0                   ; io_fstrg
        moveq   #$40,d2                 ; this much data
        moveq   #0,d3                   ; no waiting
        lea     -$64(a6),a1             ; into this buffer
        trap    #3
        cmp.l   #-1,d0                  ; see if error or data
        bge     gdata
        move.l  d0,$c(a4)               ; flag eof or other error
        bra     wait
gdata:
        clr.l   $c(a4)                  ; no error
        tst.l   d1                      ; any data fetched?
        beq     wait
        lea     -$64(a6),a1             ; buffer base
        move.l  4(a4),a2                ; write pointer
        move.l  8(a4),a3                ; read pointer
        move.l  (a7),a0                 ; a0 = screen
        bra     lp_ent
lp_start:
        move.b  (a1)+,(a2)              ; write byte to queue
        cmpa.l  $0(a4),a2               ; queue end?
        beq     wrap
        lea     1(a2),a2                ; next write position
retry:
 bra lp_ent
        cmpa.l  8(a4),a2                ; buffer full?
        bne     lp_ent
        movem.l d1/d3/a0-a1,-(a7)
        moveq   #8,d0                   ; suspend me
        moveq   #-1,d1
        moveq   #10,d3                  ; 1/5 second
        suba.l  a1,a1
        trap    #1
        movem.l (a7)+,d1/d3/a0-a1
        bra     retry                   ; see if anyone has made room in queue
wrap:
        lea     $10(a4),a2              ; reset to start of queue
lp_ent:
        dbra    d1,lp_start
        move.l  a2,4(a4)                ; set new write pointer
        bra     loop                    ; dont wait, just go again
wait:
        moveq   #8,d0                   ; suspend me
        moveq   #-1,d1
        moveq   #5,d3                   ; 1/10 second
        suba.l  a1,a1
        trap    #1
        bra     loop                    ; go round again
