#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include "session.h"
#include "lutil.h"
#include "ttyio.h"
#include "statetbl.h"
#include "version.h"

#define XMBLKSIZ 128
#define DEFAULT_WINDOW 6

extern unsigned short crc16(char*,int);
extern unsigned char checksum(char*,int);
extern long atol(char*);
extern time_t mtime2sl(time_t);
extern int m7send(char*);
extern chanid_t screen;
static char *ln,*rn;
static int flg;

static int xm_send(void);

int xmsend(char*,char*,int);
int xmsend(char *local,char *remote,int fl)
{
        int rc;
        ln=local;
        rn=remote;
        flg=fl;

        rc=xm_send();
        TTYWAIT(1);
        if (rc) loginf("send failed");
        return rc;
}

SM_DECL(xm_send,"xmsend")
SM_STATES
        sendm7,sendblk0,sendblk,writeblk,
        waitack,resync
SM_NAMES
        "sendm7","sendblk0","sendblk","writeblk",
        "waitack","resync"
SM_EDECL

        FILE *fp;
        struct stat st;
        unsigned short lcrc=0,rcrc;
        int startstate;
        int crcmode,telink;
        int CHR,a1,a2;
        int i;
        time_t seatime;
        time_t stm,etm;

        unsigned char header=SOH;
        struct _xmblk {
                unsigned char n1;
                unsigned char n2;
                char data[XMBLKSIZ];
                unsigned char c1;
                unsigned char c2;
        } xmblk;
/* XS0/XS0T */
        int NumNAK=0;
        int ACKsRcvd=0;
        int window=1;
        int sealink=0;
        int ACKST=0;
        int last_blk=-1;
        int SendBLK=0;
        int lastsent=-1;
        int ACKBLK=-1;
        int ARBLK8=0;
        int ARBLK;
        int cancount=0;
        long tmp;
        time_t T1;

        char resynbuf[16];

        (void)time(&stm);

        T1 = time(NULL);
        T1 += 30;

        /* if we are in CRC mode, assume remote is sealink capable... */

        if (session_flags & FTSC_XMODEM_CRC)
        {
                telink=0;
                crcmode=1;
                session_flags |= FTSC_XMODEM_RES;
                session_flags |= FTSC_XMODEM_SLO;
                session_flags |= FTSC_XMODEM_XOF;
                startstate=sendblk0;
        }
        else
        {
                telink=1;
                crcmode=0;
                session_flags &= ~FTSC_XMODEM_RES;
                session_flags &= ~FTSC_XMODEM_SLO;
                session_flags |= FTSC_XMODEM_XOF;
                if (flg && !(session_flags & SESSION_IFNA))
                        startstate=sendm7;
                else startstate=sendblk0;
        }

        if (stat(ln,&st) != 0)
        {
                logerr("$cannot stat local file \"%s\" to send",S(ln));
                return 0;
        }
        last_blk=(st.st_size-1)/XMBLKSIZ+1;

        if ((fp=fopen(ln,"r")) == NULL)
        {
                logerr("$cannot open local file \"%s\" to send",S(ln));
                return 0;
        }
        loginf("%s%s send \"%s\" as \"%s\", size=%lu",
                telink?"Xmodem":"Sealink",
                ((!telink) && (session_flags & FTSC_XMODEM_SLO))?" Overdrive":"",
                S(ln),S(rn),st.st_size);

SM_START(startstate)

SM_STATE(sendm7)
        if (m7send(rn)) {SM_PROCEED(sendblk0);}
        else {SM_ERROR;}

SM_STATE(sendblk0)

        memset(xmblk.data,0,sizeof(xmblk.data));

        xmblk.data[0]=(st.st_size)&0xff;
        xmblk.data[1]=(st.st_size>>8)&0xff;
        xmblk.data[2]=(st.st_size>>16)&0xff;
        xmblk.data[3]=(st.st_size>>24)&0xff;
        seatime=mtime2sl(st.st_mtime);
        xmblk.data[4]=(seatime)&0xff;
        xmblk.data[5]=(seatime>>8)&0xff;
        xmblk.data[6]=(seatime>>16)&0xff;
        xmblk.data[7]=(seatime>>24)&0xff;
        strncpy(xmblk.data+8,rn,17);
        if (telink)
                for (i=23;(i>8) && (xmblk.data[i] == '\0');i--)
                        xmblk.data[i]=' ';
        sprintf(xmblk.data+25,"ifcico %s",version);
        xmblk.data[40]=((session_flags & FTSC_XMODEM_SLO) != 0);
        xmblk.data[41]=((session_flags & FTSC_XMODEM_RES) != 0);
        xmblk.data[42]=((session_flags & FTSC_XMODEM_XOF) != 0);
        if(telink)
           debug(19,"telink block: \"%s\"",printable(xmblk.data,44));
        else
           debug(19,"sealink block: \"%s\"",printable(xmblk.data,44));
/* XS1 */
        SM_PROCEED(waitack);

SM_STATE(sendblk)
/* XS2 */

/* XS2.1 */     if ((NumNAK > 4) && (SendBLK == 0))
                {
                  debug(20,"XS2.1");
                  if (sealink)
                  {
                    loginf("Sealink header not recognised. Trying Telink");
                    sealink = 0;
                    telink = 1;
                    {SM_PROCEED(sendblk0);}
                  }
                  else if (telink)
                  {
                    loginf("Telink header not recognised. Trying Xmodem");
                    NumNAK = 0;
                    ACKBLK++;
                    SendBLK++;
                  }
                }
                if (SendBLK == 0)
                {
                   SM_PROCEED(writeblk);
                }
/* XS2.2 */     if (NumNAK > 10)
                {
                   loginf("Too many NAKs");
                   SM_ERROR;
                }
/* XS2.3 */     if (time(NULL) > T1)
                {
                   loginf("Fatal timeout");
                   SM_ERROR;
                }
/* XS2.4 */     if (SendBLK > last_blk+1)
                {
/* XS3.1 */        if (ACKBLK < (last_blk+1))
                      {SM_PROCEED(waitack);}
/* XS3.2 */        if (ACKBLK == (last_blk+1))
                   {
                        (void)time(&etm);
                        if (etm == stm) etm++;
                        loginf("sent %lu bytes in %lu seconds (%lu cps)",
                                st.st_size,etm-stm,st.st_size/(etm-stm));
                        fclose(fp);
                        SM_SUCCESS;
                   }
                }
/* XS2.5 */     if (SendBLK > (ACKBLK+window))
                {
                    SM_PROCEED(waitack);
                }
/* XS2.6 */     if (SendBLK == (last_blk+1))
                {
                   debug(20,"XS2.6");
                   PUTCHAR(EOT);
                   SendBLK++;
                   T1 = time(NULL);
                   T1 += 30;
                   SM_PROCEED(waitack);
                }
/* XS2.7 */
        debug(19,"xmsendblk send:%ld, ackd:%ld, last:%ld",
                SendBLK,ACKBLK,last_blk);

        memset(xmblk.data,SUB,sizeof(xmblk.data));


        if ((SendBLK > 0) && (lastsent != (SendBLK-1)))
        {
          if (fseek(fp,(SendBLK-1)*XMBLKSIZ,SEEK_SET) != 0)
          {
                logerr("$error seeking block %ld (byte %lu) in file \"%s\"",
                        SendBLK,(SendBLK-1)*XMBLKSIZ,S(ln));
                SM_ERROR;
          }
          else
          {
                debug(19,"Seek to block %ld (byte %lu) in file \"%s\"",
                        SendBLK,(SendBLK-1)*XMBLKSIZ,S(ln));
          }
        }
        if (fread(xmblk.data,1,XMBLKSIZ,fp) <= 0)
        {
                logerr("$read error for block %lu in file \"%s\"",
                        SendBLK,S(ln));
                SM_ERROR;
        }
        SM_PROCEED(writeblk);

SM_STATE(writeblk)

        xmblk.n1=SendBLK&0xff;
        xmblk.n2=~xmblk.n1;
        if (crcmode)
        {
                lcrc=crc16(xmblk.data,sizeof(xmblk.data));
                xmblk.c1=(lcrc>>8)&0xff;
                xmblk.c2=lcrc&0xff;
        }
        else
        {
                xmblk.c1=checksum(xmblk.data,sizeof(xmblk.data));
        }
        
        PUTCHAR(header);
        PUT((char*)&xmblk,crcmode?sizeof(xmblk):sizeof(xmblk)-1);
        if (STATUS) {SM_ERROR;}
        if (crcmode)
                debug(19,"sent 0x%02x %d bytes crc 0x%04x",
                        xmblk.n1,XMBLKSIZ,lcrc);
        else
                debug(19,"sent 0x%02x, %d bytes checksum 0x%02x",
                        xmblk.n1,XMBLKSIZ,xmblk.c1);
/* XS2.7/8 */
        T1 = time(NULL);
        if ((sealink) && ((session_flags & FTSC_XMODEM_SLO) != 0))
        {
           ACKBLK = SendBLK;
           T1 += 60;
        }
        else
           T1 += 30;

        lastsent = SendBLK;
        SendBLK++;
        SM_PROCEED(waitack);

SM_STATE(waitack)
/* XS1 */
AC0:
/* AC0.1/2 */

        TTYWAIT(0);
        CHR=GETCHAR(20);
        if (CHR == EMPTY)
           {SM_PROCEED(sendblk);}
        if (CHR == HANGUP)
            {SM_ERROR;}
        TTYWAIT(1);
        if ((CHR < 32) || (CHR > 127))
           debug(20,"AC0 got \\%02x, %d",CHR,ACKST);
        else
           debug(20,"AC0 got \"%s\", %d",printablec(CHR),ACKST);

/* AC1 */
        if (ACKST > 2)
           goto AC2;
        goto AC6;
AC2:
        if (ARBLK8 == ((~CHR) & 0xff))
        {
           ARBLK = SendBLK - ((SendBLK - ARBLK8) & 0xff);
           goto AC3;
        }
        sealink = 0;
        window = 1;
        ACKST = 0;
        goto AC6;

AC3:
        if (( ARBLK < 0) ||
            ( ARBLK > SendBLK) ||
            ( ARBLK <= (SendBLK - 128)))
        {SM_PROCEED(waitack);}

        if (ACKST == 3)
        {
           debug(19,"ACK %d, sent %d",ARBLK,SendBLK);
           sealink = 1;
           window = DEFAULT_WINDOW;
           ACKBLK = ARBLK;
           ACKsRcvd++;
           ACKST = 0;
           goto AC5;
        }

        debug(19,"NAK %d, sent %d",ARBLK,SendBLK);
        SendBLK = ARBLK;
        ACKST = 0;
        /* fall through to AC4 */
AC4:
        if (NumNAK < 4)
        {
          sealink = 1;
          window = DEFAULT_WINDOW;
        }
        else
        {
          sealink = 0;
          window = 1;
        }
        SM_PROCEED(sendblk);
AC5:
        if (((session_flags & FTSC_XMODEM_SLO) != 0) && (ACKsRcvd > 9))
        {
           session_flags &= ~FTSC_XMODEM_SLO;
           loginf("Overdrive disengaged");
        }
        SM_PROCEED(sendblk);

AC6:
        if ((ACKST==1) || (ACKST==2))
        {
           ARBLK8 = CHR;
           ACKST += 2;
           /* fall through */
        }
/* AC6.2/3 */
        if ((sealink == 0) || (ACKST == 0))
        {
/* AC7 */
           switch(CHR)
           {
           case ACK:    ACKST = 1;
                        NumNAK = 0;
                        goto AC8;

           case CRC:    crcmode = 1;
                /* fall through */

           case NAK:    ACKST = 2;
                        NumNAK++;
                        sleep(1);
                        goto AC9;

           case SYN:    goto AC10;

           case DC3:    if (session_flags & FTSC_XMODEM_XOF)
                        loginf("XOF");
                        while (((CHR=GETCHAR(15)) > 0) && (CHR != DC1))
                                if (CHR < 0) debug(19,"got %d waiting for DC1",CHR);
                                else debug(19,"got '%s' waiting for DC1",
                                        printablec(CHR));
                        loginf("XON");
                        goto ACexit;

           case CAN:    if (cancount++ > 5)
                        {
                                loginf("remote requested cancel transfer");
                                SM_ERROR;
                        }
                        break;

           default:     break;
           }
           goto AC0;
        }
ACexit:
        SM_PROCEED(sendblk);

AC8:
        if (sealink == 0)
           ACKBLK++;
        SM_PROCEED(waitack);

AC9:
        if (sealink == 0)
          SendBLK = ACKBLK+1;
        SM_PROCEED(waitack);

SM_STATE(resync)
AC10:
/* AC10.1 */
        if (NumNAK++ > 9)
        {
                loginf("too may tries to resync");
                SM_ERROR;
        }

        i=-1;
        do
        {
                debug(20,"AC10.1");
                CHR=GETCHAR(15);
                resynbuf[++i]=CHR;
        }
        while ((CHR >= '0') && (CHR <= '9') && (i < sizeof(resynbuf)-1));
        resynbuf[i]='\0';
        debug(19,"got resync \"%s\", i=%d",resynbuf,i);
        lcrc=crc16(resynbuf,strlen(resynbuf));
        rcrc=0;
        if (CHR != ETX)
        {
                if (CHR > 0) loginf("got %d waiting for resync",CHR);
                else loginf("got %s waiting for resync",printablec(CHR));
                PUTCHAR(NAK);
                SM_PROCEED(waitack);
        }
        debug(20,"AC10.1b");
        if ((CHR=GETCHAR(1)) < 0)
        {
                PUTCHAR(NAK);
                SM_PROCEED(waitack);
        }
        rcrc=CHR&0xff;
        debug(20,"AC10.1c");
        if ((CHR=GETCHAR(1)) < 0)
        {
                PUTCHAR(NAK);
                SM_PROCEED(waitack);
        }
        rcrc |= (CHR << 8);
        if (rcrc != lcrc)
        {
                loginf("bad resync crc: 0x%04x != 0x%04x",lcrc,rcrc);
                PUTCHAR(NAK);
                SM_PROCEED(waitack);
        }
/* AC10.2 */
        sealink=1;
        window=DEFAULT_WINDOW;
        SendBLK=atol(resynbuf);
        ACKBLK=SendBLK-1;
        loginf("resyncing at block %ld (byte %lu)",
                SendBLK,(SendBLK-1L)*XMBLKSIZ);
        PUTCHAR(ACK);
        SM_PROCEED(sendblk);

SM_END
SM_RETURN


int xmsndfiles(file_list*);
int xmsndfiles(file_list *tosend)
{
        int rc,c,gotnak,count;
        file_list *nextsend;

        for (nextsend=tosend;nextsend;nextsend=nextsend->next)
        if (*(nextsend->local) != '~')
        {
                if (nextsend->remote)
                if ((rc=xmsend(nextsend->local,nextsend->remote,
                                (nextsend != tosend)))) /* send m7 for rest */
                        return rc; /* and thus avoid execute_disposition() */
                else
                {
                        gotnak=0;
                        count=0;
                        while (!gotnak && (count < 6))
                        {
                                debug(20,"waitNAK");
                                c=GETCHAR(15);
                                if (c < 0) return STATUS;
                                if (c == CAN)
                                {
                                        loginf("remote refused receiving");
                                        return 1;
                                }
                                if ((c == CRC) || (c == NAK)) gotnak=1;
                                else debug(19,"got '%s' waiting NAK",
                                        printablec(c));
                        }
                        if (c == CRC) session_flags |= FTSC_XMODEM_CRC;
                        if (!gotnak) return 1;
                }
                execute_disposition(nextsend);
        }
        PUTCHAR(EOT);
        return STATUS;
}
